#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <strings.h>
#include <sys/soundcard.h>
#include <math.h>

/*
	This is the general purpose program invoked by the ans script to
	play files to the phone, record sound from the phone into a file,
	and recognize caller touchtones via FFT (see fftlib.c).

    Copyright (c) 1999, Rich Drewes
*/

/*
	1) If you compute the standard deviation for the energy values
	across the entire spectrum, and if the entire spectrum is
	much larger than the bandwidth limited phone signal (e.g. 16K
	samples on the 30-3000Hz phone line input) you may end up with
	an artificially low stddev value since you have a large region
	from 3kHz to 8kHz that is essentially 0 (phone bandwidth limited
	input).

	So:  if bandwidth of grab is much greater than bandwidth of phone
	we simply only do the stddev algorithm up to 3.5kHz, which is about
	the maximum bandwidth we can get over the phone.

	Then it shouldn't matter whether we do 8K or 16K samples.

	2)  probably it would be better to do a 'moving window' FFT rather
	than 'get 1024 samples, do FFT on *that*, get next 1024 samples,
	do FFT on *that* block . . . ' this would give better temporal
	resolution on recognitions.  However, if the fft size translates
	into fairly small real time value (if sample rate is high enough)
	then it's probably OK to do it the simple 'get x samples, do fft,
	get next x, do fft . . . "

	----
	Ignore this:

    256 samples at
	8KHz takes 32 ms to gather, 1 sample is 125 us
	16KHz takes 16 ms to gather, 1 sample is 62.5 us

    perf (on old P166?):
	did 1000 256 pt ffts in 13 secs, 1 per 13 ms
	did 1000 512 pt ffts in 30 secs, 1 per 30 ms

	----
	Ignore this too:

	say we do a 512 point fft on 16kHz sample stream
	our fft algorithm puts results in 0..(512-1) elements in array
	first point:  the highest frequency we can resolve is 1/2 of the sampling rate, or 8KHz.  after that we get aliasing
	the first 256 results of our fft give us 8KHz of resolution.  the 0th and (512-1) sample is DC, and the 255 (and 256) is 8KHz

	so the the freq[i] = (8k/255)*i

	----
	Touchtone DTMF frequencies:

        1209 1336 1477 1633
 
     697   1    2    3    A

     770   4    5    6    B

     852   7    8    9    C

     941   *    0    #    D

	----
Dial Tone         350 & 440     ---     ---
Busy Signal       480 & 620     0.5     0.5
Toll Congestion   480 & 620     0.2     0.3
Ringback (Normal) 440 & 480     2.0     4.0
Ringback (PBX)    440 & 480     1.5     4.5
Reorder (Local)   480 & 620     3.0     2.0
Invalid Number    200 & 400
Hang Up Warning 1400 & 2060     0.1     0.1
Hang Up         2450 & 2600     ---     ---

	----

	frequencies double each octave

	From web:
	"middle C is 523.25 Hertz"
	"A3" (the note "A" at the third octave) is deemed to be at 440 Hz and, therefore, "A4" will be at 880 Hz.

   Calculation for Equal-Tempered tuning [A3 = 440Hz]
Hertz      Octave=0    Octave=1    Octave=2    Octave=3    Octave=4     Octave=5

 0     A     55.000     110.000     220.000     440.000     880.000    1,760.000
 1   A#/Bb   58.270     116.541     233.082     466.164     932.328    1,864.655
 2     B     61.735     123.471     246.942     493.883     987.767    1,975.533
 3     C     65.406     130.813     261.626     523.251   1,046.502    2,093.005
 4   C#/Db   69.296     138.591     277.183     554.365   1,108.731    2,217.461
 5     D     73.416     146.832     293.665     587.330   1,174.659    2,349.318
 6   D#/Eb   77.782     155.563     311.127     622.254   1,244.508    2,489.016
 7     E     82.407     164.814     329.628     659.255   1,318.510    2,637.020
 8     F     87.307     174.614     349.228     698.456   1,396.913    2,793.826
 9   F#/Gb   92.499     184.997     369.994     739.989   1,479.978    2,959.955
10     G     97.999     195.998     391.995     783.991   1,567.982    3,135.963
11   G#/Ab  103.826     207.652     415.305     830.609   1,661.219    3,322.438
12     A    110.000     220.000     440.000     880.000   1,760.000    3,520.000
	
	----

    on new P2 333Mhz:
	10000 ffts in 48 secs, each took 0.004800 secs, 16 bit samples, 16000 samples/second 

    why no sound record overrun (or very few) at 8 bits 16KHz?
    it takes us 30 ms to process 512 bytes of data, which
    roughly corresponds to 32 ms worth of realtime data . . . 

    same holds true at 16 bit samples.  why?  it doesn't take us
    any longer to process it since we convert everything to float
    anyway.  the internal buffer will fill up faster with 16 bits
    per sample, but since we read a fixed number of samples each
    fft anyway, we read twice as many bytes at 16 bits per sample

    get sound record overruns when I change from window to window since
    that takes extra cpu time we don't have to spare!    

	Notes:
		according to VoxWare manual, 8 bit samples are centered around 128.
		these can be scaled by subtracting 128 from them or
			equivalently XORing with 0x80
		however 16 bit samples come in as *signed shorts*!
*/

char *filetoplay;
int playfile=0;
int playfilefd=-1;
int analyzeplay=0;
int computedenergy=0;

char *filetowrite;
int writefile=0;
int writefilefd=-1;
float writetime=0.0;

#define MAXTONES 80
int gettones=0;
char toneoutput[MAXTONES];
int ontone=0;
int tonestoget=0;

int fftpoints;
int fftbits;

int mode=0;
int printdynrange;
float fftrealenergy(float*, int);
//float fftfindpeaks(float*, float*, int);
float fftprintenergy(float*, float*, int);
int fftinit(int);
int fft(float*, float*, int, int);
int NextFFT(unsigned char *buf);
float *xr, *xi;
float *en;					// square of energy at freq., square of real plus square of imag.  NOT SQUARE ROOTED THOUGH
//float *smoothen;			// smoothed energy; average of neigboring components

#define DEFAULT_DSP_SPEED 8000

#define RECORD	0
#define PLAY	1
#define AUDIO "/dev/dsp"

int timelimit = 0, dsp_speed = DEFAULT_DSP_SPEED, dsp_stereo = 0;
int curcol=0;
int curcolenergy=0;
int samplesize = 8;
int quiet_mode = 0;
int audio, abuf_size;
int direction, omode;
char *audiobuf, c;

void recplay ();

#ifdef DOXPLOT
void PlotLine(int color, int x, int y, int x1, int y1);
#endif

int istep, a3pos, a4pos, a5pos;
float fstep, secperblock;

#if 0
#define READSAMPLES	2048
#define READBITS	11
#endif
#if 0
#define READSAMPLES	1024
#define READBITS	10
#endif
#if 0
#define READSAMPLES	512
#define READBITS	9
#endif
#if 0
#define READSAMPLES	256
#define READBITS	8
#endif
#if 0
#define READSAMPLES	128
#define READBITS	7
#endif
#if 0
#define READSAMPLES	64
#define READBITS	6
#endif
#if 0
#define READSAMPLES	32
#define READBITS	5
#endif

typedef struct note {
	float freqhz;
	char *name;
} Note;

Note nf[]= { {55.000, "A"}, {58.270, "A#/Bb"}, {61.735, "B"}, {65.406, "C"},
			 {69.296, "C#/Db"}, {73.416, "D"}, {77.782, "D#/Eb"}, {82.407, "E"},
			 {87.307, "F"}, {92.499, "F#/Gb"}, {97.999, "G"}, {103.826, "G#/Ab"}, {0.0, NULL} };

typedef struct dtmf {
	int freqhz1;
	int freqhz2;
	char *name;
	int set;
	int tmp;
	float ontime;
} Dtmf;

#define MAXDETECTTONE	15		// don't want dialtone or busy to be detected as input tone	
Dtmf dtones[]= { {1336, 941, "0",}, {1209, 697, "1"}, {1336, 697, "2"}, {1477, 697, "3"},
				 					{1209, 770, "4"}, {1336, 770, "5"}, {1477, 770, "6"},
									{1209, 852, "7"}, {1336, 852, "8"}, {1477, 852, "9"},
				 {1633, 698, "A"},  {1633, 941, "B"}, {1633, 941, "C"}, {1633, 941, "D"},
				 {1209, 941, "*"},  {1477, 941, "#"}, {350, 440, "DIALTONE"}, {480, 620, "BUSY"},
				 {0, 0, NULL} };


#define NUMOCTAVES	5

ClearDtones()
{
	int i=0;

	fprintf(stderr, "enter ClearDtones\n"); fflush(stdout);
	while(dtones[i].name) {
		dtones[i].set=0;
		dtones[i].ontime=0.0;
		i++;
	}
	fprintf(stderr, "exit ClearDtones\n"); fflush(stdout);
}

FreqToNote(int freq)
{
	int i, m=1, n;

	for(i=0; i<NUMOCTAVES; i++) {
		n=0;
		while(nf[n].name) { 
			float recfreq=nf[n].freqhz * (float) m;
			if(fabs(recfreq - (float) freq) < 10.0) {
				fprintf(stderr, "Got note %s, octave %d, freq %f (actual %d)\n",  nf[n].name, i, recfreq, freq);
				return;
			}
			n++;
		}
		m*=2;	// go to next octave
	}
}

// following is experimental code for trying to decode caller ID Bell 202 modem tone
AnalyzePlay()
{
	float windowsize=(1.0 / 600.0) / 1.0;	// in seconds, .001 = 1ms
	off_t fsize;
	char *fdata;

	if(playfile) {
		int winstart=0;		// starting index of current sliding window
		int windownum=0;
		int nsamps;
		float sampsperwin;

		playfilefd=open(filetoplay, O_RDONLY);
		if(playfilefd < 0) {
			fprintf(stderr, "failed open file %s\n", filetoplay);
			exit(-1);
		}

		if(samplesize != 16) {
			printf("analyze is currently only supported for 16 bit input samples\n");
			exit(-1);
		}

		fsize=lseek(playfilefd, 0, SEEK_END);
		lseek(playfilefd, 0, SEEK_SET);
		printf("file is %d bytes\n", (int) fsize);
		fdata=(char *) malloc((int) fsize);
		read(playfilefd, fdata, (int) fsize);
		printf("read file into buffer\n");
		nsamps = (int) fsize / 2; 

		{
			int i;
			short *samp=(short *) fdata;
			for(i=0; i<nsamps; i++) {
				printf("%d ", *samp);
				samp++;
			}
		}

		sampsperwin=(float) dsp_speed * (float) windowsize;
		printf("window size is %f seconds, %d samples\n", windowsize, (int) sampsperwin);
		printf("samplesize in bits:  %d, sample rate %d, sample block size %d\n", samplesize, dsp_speed, fftpoints);
		printf("window is %f samples, the fft uses %d samples\n", sampsperwin, fftpoints);

		// rather than doing fft's for each block of samples we do moving window fft's each x ms

		while((winstart + fftpoints) < nsamps) {
			int i, peaks[8];
			short *samp;
			float secs;

			samp=(short *) ((short *) fdata + winstart);
			secs=(float) winstart/(float) dsp_speed;

			fprintf(stderr, "window number %d, starting at index %d, which is %f s, starting ptr %0x (%d)\n", windownum, winstart, secs, samp, samp);

			for(i=0; i<fftpoints; i++) {
				//xr[i] = (float) *samp; xi[i] = 0.0;
				// for high frequency resolution, want to do a many point fft.  however, we want the windows size
				// to be small (to get good time resolution) BUT
				// the number of samples we will need to fill up a large fft is relatively large and will encompass several time
				// steps, which is undesirable.  therefore we will not replicate the same data again and again within the window
				// in order to fill it up and not introduce data from the next samples
				// this may introduce low frequency and high frequency garbage output from the fft, but should increase time discrimination
				xr[i] = (float) samp[i%(int) sampsperwin]; xi[i] = 0.0;
				//samp++;
			}
			fft(xr, xi, fftbits, 1);
			ComputeEnergy(xr, xi, en, fftpoints);
			FindPeaks(en, peaks, 8);
			printf("%f s:  ", secs); DumpPeaks(peaks, 8);
			windownum++;
			winstart=(int) ((float) windownum * sampsperwin);
		}

		fflush(stdout);
		close(playfilefd);
	}
}

int main(int argc, char *argv[])
{

	char *command;
	int i;
	int tmp;

	toneoutput[0]='\0';
	direction = RECORD;
	//omode = O_RDWR|O_NDELAY;
	omode = O_RDWR;

#ifdef DOXPLOT
	PlotInitialize();
	/*PlotLine(1, 0, 0, 100, 100);*/
#endif

	fftpoints=1024;
	fftbits=10;

	ParseCmdLine(argc, argv);

	xr=(float *) malloc(sizeof(float) * fftpoints);
	xi=(float *) malloc(sizeof(float) * fftpoints);
	en=(float *) malloc(sizeof(float) * fftpoints);

	ClearDtones();
	
	audio = open(AUDIO, omode, 0);
	if(audio == -1){
		perror(AUDIO);
        exit(-1);
	}

#if 1
	// note:  this fragment isn't necessarily the size we are reading (fftsize). 
	// right now the only time we query the driver using this is when we are playing
	// a file, and in that case we are dealing with WBUFSIZE buffers
	// when we read a block of fftsize from the audio device (mic input) we don't
	// check with the driver to see if a block is there.  we really should check there
	// too since otherwise we could create starvation problems playing a file if
	// the fftsize is large relative to the WBUFSIZE . . . 
	#define SNDFRAGVAL      0x0010000b;                     // 2^b = 2^11 = 2048
	#define WBUFSIZE	2048
	{
	int arg=SNDFRAGVAL
	if(ioctl(audio, SNDCTL_DSP_SETFRAGMENT, &arg) < 0) {
		fprintf(stderr, "can't set frag size; no sound\n");
		exit(-1);
	}
	}
#endif

	tmp = samplesize;
	ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &samplesize);
	if(tmp != samplesize){
		fprintf(stderr, "Unable to set the sample size\n");
		exit(-1);
	}

	if(ioctl(audio, SNDCTL_DSP_STEREO, &dsp_stereo)==-1){
		fprintf (stderr, "%s: Unable to set mono/stereo\n", command);
		perror (AUDIO);
		exit(-1);
	}

	if(ioctl(audio, SNDCTL_DSP_SPEED, &dsp_speed) == -1){
		fprintf (stderr, "%s: Unable to set audio speed\n", command);
		perror(AUDIO);
		exit(-1);
	}

	if(ioctl(audio, SNDCTL_DSP_SYNC, NULL) < 0) {
		fprintf(stderr, "can't sync dsp; turning sound off\n");
		exit(-1);
	}

	ioctl(audio, SNDCTL_DSP_GETBLKSIZE, &abuf_size);
	if(abuf_size < 1024 || abuf_size > (2*65536)){
		if(abuf_size == -1) {
			perror(AUDIO);
		} else {
			fprintf(stderr, "Invalid audio buffers size %d\n", abuf_size);
      		exit(-1);
    	}
	}

	if ((audiobuf = malloc (abuf_size)) == NULL){
		fprintf(stderr, "Unable to allocate input/output buffer\n");
		exit(-1);
	}

	for(i=0; i<fftpoints; i++) {
		xr[i]=0.0;
  	}

	if(fftinit(fftbits) < 0) {
		fprintf(stderr, "fftinit failed\n");
		exit(-1);
  	}

//	fstep=(float) dsp_speed/(float) ((fftpoints/2)-1);
	fstep=(float) dsp_speed/(float) fftpoints;
	secperblock=(float) fftpoints / (float) dsp_speed;
	a3pos=(int) ((float) 440.0 / fstep);
	a4pos=(int) ((float) 880.0 / fstep);
	a5pos=(int) ((float) 1760.0 / fstep);
	fprintf(stderr, "fstep is %f--each pixel on x axis represents %f Hz, A3 is at x position %d\n", fstep, fstep, a3pos);

#ifdef DOXPLOT
	PlotLine(2, a3pos, -100, a3pos, -50);
	PlotLine(2, a4pos, -100, a4pos, -50);
	PlotLine(2, a5pos, -100, a5pos, -50);
#endif

	fprintf(stderr, "fft points:  %d, bits %d\n", fftpoints, fftbits);

	if(analyzeplay) {
		AnalyzePlay();
	}

	if(playfile) {
		playfilefd=open(filetoplay, O_RDONLY);
		if(playfilefd < 0) {
			fprintf(stderr, "failed open file %s\n", filetoplay);
			exit(-1);
		}
	}
	if(writefile) {
		writefilefd=open(filetowrite, O_CREAT|O_TRUNC|O_WRONLY, 0666);
		if(writefilefd < 0) {
			fprintf(stderr, "failed open file %s\n", filetowrite);
			exit(-1);
		}
	}
	recplay();

	close (audio);
	exit(0);
}

Exit(int n)
{
	close(audio);
	printf("%s\n", toneoutput);
	exit(n);
}

DumpPeaks(int peaks[], int npeaks)
{
	int i=0;

	printf("peaks:  ");
	while(i<npeaks) {
		if(peaks[i] < 0) break;
		printf("%d ", peaks[i]);
		i++;
	}
	printf("\n");
}

// return Hz value for fft index i
int PosToHz(int i)
{
	return ((float) i * fstep);
}

int HzToPos(int hz)
{
	return (int) ((float) hz / fstep);
}

void recplay()
{
	int fd, l;

	int count=0, c, startt=time(0), endt, numc=1000, elt;

	fprintf(stderr, "entering main loop, samplesize in bits:  %d, sample rate %d, sample block size %d\n", samplesize, dsp_speed, fftpoints);

	while(1) {
		computedenergy=0;

		// do a write before ever doing a read, so we don't get initial pop for lack of data
		if(playfile) WriteSound();

		// now process input from microphone
		c = fftpoints*(samplesize/8);

		if((l = read(audio, audiobuf, c)) > 0){
			NextFFT(audiobuf);
	    }

		if(gettones) {
			ComputeEnergy(xr, xi, en, fftpoints);
			FindPeaks(en, NULL, 0);
		}

		if(writefile) {
			write(writefilefd, audiobuf, c);
			writetime -= (float) fftpoints / (float) dsp_speed;
			fprintf(stderr, "time remaining to write:  %f\n", writetime);
			if(writetime <= 0.0) {
				writefile=0;
				close(writefilefd);
				Exit(0);
			}
		}

		if(l==-1) {
			perror(AUDIO);
			exit(-1);
	    }
		#if 0
		// if we opened audio device with O_NDELAY, we would need to verify that we actually got 
		if(l==c) {
		}
		#endif
		count++;
		if(count%numc == 0) {
			printdynrange=1;
		} else {
			printdynrange=0;
		}

		if((count%numc) == 0) {
			endt=time(0);
			if((elt=(endt-startt)) == 0) continue;
			fprintf(stderr, "%d blocks in %d secs, %f secs for each %d sample block\n", count, elt, (float) elt/ (float) count, fftpoints);
			fprintf(stderr, "%f samples per second\n", (float) (count*fftpoints)/(float) elt);
		}

	}
}

// read a buffer of size WBUFSIZE from the file we are playing and write it to the audio
// device, IFF the driver tells us that we can write without blocking
WriteSound()
{
	int res;
	audio_buf_info info;
	char wbuf[WBUFSIZE];

//	fprintf(stderr, "in WriteSound\n");

	ioctl(audio, SNDCTL_DSP_GETOSPACE, &info);
#if 0
	fprintf(stderr, "%d fragments write without block, %d total fragments, %d bytes in fragment, %d bytes write without block\n", info.fragments,
		info.fragstotal, info.fragsize, info.bytes);
#endif

	if(info.fragments < 1) {		// can't write another buffer without blocking.  so don't try!
//		fprintf(stderr, "fragments is < 1 (%d)\n", info.fragments);
		return;
	} 

	res=read(playfilefd, wbuf, WBUFSIZE);

//	fprintf(stderr, "read %d bytes\n", res);

	if(res == WBUFSIZE) {
		write(audio, wbuf, res);
		// how will we know when audio has finished playing?
	} else {
		write(audio, wbuf, res);
		// set playfile to done . . . 
		playfile=0;		// we won't need to write anymore data . . . 
		fprintf(stderr, "turned off playfile, done writing to audio device\n");
		Exit(0);
	}
}

// compute energy^2 (sum of squares of real and imaginary parts) for first half of nsamp points
ComputeEnergy(float xr[], float xi[], float en[], int nsamp)
{
	int i;

	if(computedenergy) return;

	for(i=0; i<(nsamp/2); i++) {
		en[i]=(xr[i]*xr[i] + xi[i]*xi[i]);
	}

	computedenergy=1;	// so we know we did it, and won't waste time doing it again
}

// x axis:  frequency, y axis:  energy at that frequency
DoPlotEnergy(float xr[], float xi[])
{
#ifdef DOXPLOT
	int i,j;
	int maxx=PlotGetMaxx();
	int step=fftpoints/maxx;
	int maxy=PlotGetMaxy();
	float energy;
	int intenergy;


	if (step < 1) step=3;		// step is how wide an area over which to perform running average
//	fprintf(stderr, "PlotEnergy:  %d samples, step %d\n", fftpoints, step);

	for(i=0; i<(fftpoints/2); i++) {
		en[i]=(xr[i]*xr[i] + xi[i]*xi[i]);
	}

	// coerce DC component to 0 to eliminate bias?

	for(i=(step +1); i<(fftpoints/2)-step; i+=1) {		// start at step+1=2 so no sample includes DC component;
		energy=en[i];
		intenergy=(int) (energy / 100.0);
		if(intenergy > maxy) intenergy=maxy;
// fix:  clear whole area with rectangle primitive rather than drawing line in background color . . . 
		PlotLine(0, i, 0, i, maxy);
		PlotLine(2, i, 0, i, intenergy);
		if(i >= maxx) { fprintf(stderr, "screen not wide enough to show full spectrum\n"); break; }
	}
	PlotSync();
#endif
}

// find first npeaks peaks (>3 stddev over mean) and put freq (in Hz) of same into p[n] 
// if can't find sufficient peaks, fill p[n] with -1; 
// also does DTMF recognition, returns results in toneoutput array
FindPeaks(float en[], int peaks[], int npeaks)
{
    int i, nsamp, endsamp, onpeak=0;
    float   max=0.0;
    float v, sum, avg, thresh, sumsq, d, stdd, stdd3, stdd2;
    int maxindex=0, pos1, pos2;
	float en1, en2;
    //float peaks[FFTPOINTS+1];

	// compute standard deviation.  don't include DC bias term (the 0th element in array)
	// stop computing above 3.5kHz since phone is bandwidth limited there anyway

	endsamp=3500.0 / (float) fstep; 	// each step is fstep Hz . . .
//	fprintf(stderr, "endpoint is element %d, would have been %d\n", endsamp, (fftpoints/2));
	nsamp=endsamp-1;		// skipping 0 element
    sum=0.0;

    for(i=1; i<endsamp; i++) {
		sum+=en[i];
	}

	avg=sum/(float) nsamp;

//	fprintf(stderr, "avg is %f\n", avg);
	sumsq=0.0;
    for(i=1; i<endsamp; i++) {
		d=en[i]-avg;
		sumsq+=d*d;
	}

	stdd=sqrt(sumsq/(float) (nsamp-1));

	stdd2=2.0*stdd;
	stdd3=3.0*stdd;

#define NOISEREJECT	10.0
//	fprintf(stderr, "stddev for energy is %f\n", stdd);
    for(i=1; i<nsamp; i++) {
		if(((en[i]-avg) > stdd3) && (avg > NOISEREJECT)) {
			int freq=PosToHz(i);

			fprintf(stderr, "position %d (%d Hz) is > 3 stddev over mean\n", i, PosToHz(i));
			if(onpeak < npeaks) {
				peaks[onpeak]=freq;
				onpeak++;
			}
			FreqToNote(freq);
		}
	}

	if(onpeak < npeaks) peaks[onpeak]=-1;	// end marker if can't find npeaks peaks

	i=0;

	// now try to recognize dual tones (DTMF)
#define DTMFNOISEREJECT	5.0

	while(dtones[i].name) {
		pos1=HzToPos(dtones[i].freqhz1);
		pos2=HzToPos(dtones[i].freqhz2);
		// put in array bound checks on pos1 & pos2
		// allow some frequency variation (and also allow for imprecision of frequency steps since
		// if we do a 1024 point fft on 8KHz, our frequency steps are about 8Hz and we could round
		// and end up looking at a neighboring sample and miss a lot of the energy . . . 
		en1=(en[pos1]-avg) + (en[pos1-1]-avg) + en[pos1+1]-avg;
		en2=(en[pos2]-avg) + (en[pos2-1]-avg) + en[pos2+1]-avg;
		en1/=3.0;
		en2/=3.0;

		if((en1 > stdd3) && (en2 > stdd3) && (en1 > DTMFNOISEREJECT) && (en2 > DTMFNOISEREJECT)) {
//			printf("detected %s (avg %f stdd %f en1 %f en2 %f)\n", dtones[i].name, avg, stdd, en[pos1], en[pos2]);
			if(!dtones[i].set) {
				dtones[i].set=1;
				dtones[i].ontime=secperblock;
//				printf("setting touchtone %s\n", dtones[i].name);
			} else {
				dtones[i].ontime+=secperblock;
//				printf("added %f to set time, now %f\n", secperblock, dtones[i].ontime);
			}
		} else if(dtones[i].set) {			// tone was pressed previously, but isn't anymore
			fprintf(stderr, "GOT TOUCHTONE %s, duration %f\n", dtones[i].name, dtones[i].ontime);
			dtones[i].set=0;
			dtones[i].ontime=0.0;
			if(i <= MAXDETECTTONE) {
				toneoutput[ontone]=dtones[i].name[0];
				ontone++;
				toneoutput[ontone]='\0';
				if(ontone >= tonestoget) {
					Exit(0);
				}
			}
		}
		i++;
	}
}

DoPlot(float xr[])
{
    int i, end=fftpoints;
    float gg;

/*fprintf(stderr, "start at %d\n", curcol);*/
/*fprintf(stderr, "end at %d\n", end);*/
#define SCALE	200.0
    if(samplesize==16) {
	for(i=0; i<end; i++) {
#ifdef DOXPLOT
	    PlotLine(0, curcol, -400, curcol, 400);
#endif
	    gg=xr[i];
	    gg=gg/SCALE;
#ifdef DOXPLOT
	    PlotLine(1, curcol, (int)gg, curcol, -(int)gg);
#endif
	    curcol++;
		/*		fprintf(stderr, "%d %f\n", i, gg);*/
	}
#ifdef DOXPLOT
		PlotSync();/**/
#endif
    } else if(samplesize==8) {
		for(i=0; i<end; i++) {
			gg=xr[i];
#ifdef DOXPLOT
			PlotLine(0, curcol, -400, curcol, 400);
			PlotLine(1, curcol, (int)gg, curcol, -(int)gg);
#endif
			curcol++;
		}
#ifdef DOXPLOT
		PlotSync();
#endif
    } else {
        fprintf(stderr, "can't nextfft!  unknown samplesize\n");
		exit(0);
    }
}

NextFFT(unsigned char buf[])
{
    int i, j, startt, endt;
    short *sdp;
    float s, e, gg, maxs=-10000.0, mins=10000.0;

    if(samplesize==16) {
		sdp=(short *)buf;
		for(i=0; i<fftpoints; i++) {
			s=(float) (*sdp);		// note:  16 bit samples are signed shorts!
	    	xr[i]=s;
	    	xi[i]=0.0;

			if(s < mins) mins=s;
			if(s > maxs) maxs=s;

	    	sdp++;
		}

    } else if(samplesize==8) {
		for(i=0; i<fftpoints; i++) {

			s=((float) buf[i]) - 128.0;	// note:  8 bit samples are unsigned and must be scaled!
	    	xr[i]=s;
	    	xi[i]=0.0;
//			fprintf(stderr, "%d %f\n", buf[i], xr[i]);
//			fprintf(stderr, "%f \n", xr[i]);
			if(s < mins) mins=s;
			if(s > maxs) maxs=s;
		}
    } else {
	    fprintf(stderr, "Can't NextFFT!  Samplesize must be 8 or 16 bits.\n");
	    Exit(0);
    }

	if(printdynrange) {
		fprintf(stderr, "dynamic range for this fft:  %f %f\n", mins, maxs);
	}

    if(mode==20) DoPlot(xr);

/*    e=fftrealenergy(xr, fftpoints);*/
/*    fprintf(stderr, "%2f \n", e);*/

    if(mode==4) {
		int el;
		int numfft=10000;

		time(&startt);
 		for(j=0; j<numfft; j++) {
	    	fft(xr, xi, fftbits, 1);
		}
		time(&endt);
		el=endt-startt;
		fprintf(stderr, "%d ffts in %d secs, each took %f secs\n", numfft, el, (float) el/(float) numfft);
    }

    if(mode==35) {
		int peaks[4];

		fft(xr, xi, fftbits, 1);
		DoPlotEnergy(xr, xi);
		//FindPeaks(en, peaks, 4);
    }
}

ParseCmdLine(argc, argv)
int argc;
char *argv[];
{
    char *vec, *vec2;
    int i;

	fprintf(stderr, "default mode is plotenergy\n");
	mode=35;

    i = 0;
    while((vec = argv[i]) != NULL) {
		if(strcmp(vec, "-fft") == 0) {
			int ff=32;

			i++;
			vec=argv[i];	
			if(!vec) return;
			fftpoints=atoi(vec);
			{

				fftbits=5;

				while(ff!=fftpoints && ff<4096) {
					ff*=2;
					fftbits++;
				}
			}
			if(ff!=fftpoints) {
				fprintf(stderr, "Invalid fft point size:  %d.\n  Must be a power of 2 32<=x<=4096\n", fftpoints);
				fprintf(stderr, "ff=%d\n", ff);
				exit(-1);
			}
			fprintf(stderr, "%d point fft, %d bits:  %d\n", fftpoints, fftbits);
		}
        if(strcmp(vec, "-12000") == 0) {
			dsp_speed=12000;
        }
        if((strcmp(vec, "-44000") == 0) || (strcmp(vec, "-44100") == 0)) {
			dsp_speed=44100;
        }
        if(strcmp(vec, "-16000") == 0) {
			dsp_speed=16000;
        }
        if(strcmp(vec, "-11025") == 0) {
			dsp_speed=11025;
        }
        if(strcmp(vec, "-16") == 0) {
			samplesize=16;
        }

		// this mode stuff is old and lame.  should just have flags that can turn
		// features on and off and can be intermixed
        if(strcmp(vec, "-peak") == 0) { mode=1; }
        if(strcmp(vec, "-print") == 0) { mode=2; }
        if(strcmp(vec, "-perf") == 0) { mode=4; }
        if(strcmp(vec, "-printline") == 0) { mode=8; }
        if(strcmp(vec, "-plotspectrum") == 0) { mode=33; }
        if(strcmp(vec, "-plotfftnorm") == 0) { mode=34; }
		if(strcmp(vec, "-analyzeplay") ==0) {
			analyzeplay=1;
			printf("Will analyze file being played.\n");
		}
        if(strcmp(vec, "-plotenergy") == 0) { mode=35; }
        if(strcmp(vec, "-plotfft") == 0) { mode=30; }
        if(strcmp(vec, "-plotraw") == 0) { mode=20; }
        if((strcmp(vec, "-playfile") == 0) || (strcmp(vec, "play") == 0)) {
			i++;
			vec=argv[i];	
			if(!vec) return;
			playfile=1;
			filetoplay=vec;
			fprintf(stderr, "will play file %s\n", filetoplay);
        }
        if((strcmp(vec, "-writefile") == 0) || (strcmp(vec, "-record") == 0)) {
			i++;
			vec=argv[i];	
			if(!vec) return;
			writefile=1;
			filetowrite=vec;
			fprintf(stderr, "will write mic input to file %s\n", filetowrite);
        }
        if(strcmp(vec, "-time") == 0) {
			i++;
			vec=argv[i];	
			if(!vec) return;
			writetime=atof(vec);
			fprintf(stderr, "will try to write %f secs of data\n", writetime);	
		}
        if(strcmp(vec, "-gettones") == 0) {
			i++;
			vec=argv[i];
			if(!vec) return;
			tonestoget=atof(vec);
			fprintf(stderr, "will try to get %d tones\n", tonestoget);	
			gettones=1;
		}
        i++;
    }
}
